import numpy as np
import torch
from torch import nn as nn
from torch.nn import functional as F

from rlkit.core.util import Wrapper
from rlkit.policies.base import ExplorationPolicy, Policy
from rlkit.torch.distributions import TanhNormal
from rlkit.torch.networks import Mlp
from rlkit.torch.core import np_ify
from rlkit.torch.transformer import BERT


LOG_SIG_MAX = 0
LOG_SIG_MIN = -20


class TanhGaussianPolicy(Mlp, ExplorationPolicy):
    """
    Usage:

    ```
    policy = TanhGaussianPolicy(...)
    action, mean, log_std, _ = policy(obs)
    action, mean, log_std, _ = policy(obs, deterministic=True)
    action, mean, log_std, log_prob = policy(obs, return_log_prob=True)
    ```
    Here, mean and log_std are the mean and log_std of the Gaussian that is
    sampled from.

    If deterministic is True, action = tanh(mean).
    If return_log_prob is False (default), log_prob = None
        This is done because computing the log_prob can be a bit expensive.
    """

    def __init__(
            self,
            hidden_sizes,
            obs_dim,
            latent_dim,
            action_dim,
            std=None,
            init_w=1e-3,
            use_layer_norm=False,
            use_dropout=False,
            hidden_activation=F.leaky_relu,
            output_activation=torch.tanh,
            **kwargs
    ):
        self.save_init_params(locals())
        super().__init__(
            hidden_sizes,
            input_size=obs_dim,
            output_size=action_dim,
            init_w=init_w,
            hidden_activation=hidden_activation,
            output_activation=output_activation,
            layer_norm=use_layer_norm,
            use_dropout=use_dropout,
            **kwargs
        )
        self.latent_dim = latent_dim
        self.log_std = None
        self.std = std
        if std is None:
            last_hidden_size = obs_dim
            if len(hidden_sizes) > 0:
                last_hidden_size = hidden_sizes[-1]
            self.last_fc_log_std = nn.Linear(last_hidden_size, action_dim)
            self.last_fc_log_std.weight.data.uniform_(-init_w, init_w)
            self.last_fc_log_std.bias.data.uniform_(-init_w, init_w)
        else:
            self.log_std = np.log(std)
            assert LOG_SIG_MIN <= self.log_std <= LOG_SIG_MAX

    def get_action(self, obs, deterministic=False):
        actions = self.get_actions(obs, deterministic=deterministic)
        return actions[0, :], {}

    @torch.no_grad()
    def get_actions(self, obs, deterministic=False):
        #outputs = self.forward(obs.size()[0], obs.size()[1], obs, deterministic=deterministic)[0]
        outputs = self.forward(1, obs.size()[0], obs, deterministic=deterministic)[0]
        return np_ify(outputs)

    def forward(
            self,
            meta_size,
            batch_size,
            obs,
            reparameterize=False,
            deterministic=False,
            return_log_prob=False,
    ):
        """
        :param obs: Observation
        :param deterministic: If True, do not sample
        :param return_log_prob: If True, return a sample and its log probability
        """
        h = obs
        for i, fc in enumerate(self.fcs):
            #h = self.hidden_activation(fc(h))
            h = fc(h)
            if self.layer_norm and i < len(self.fcs) - 1:
                h = self.layer_norms[i](h)
            h = self.hidden_activation(h)
            if self.use_dropout and i < len(self.fcs) - 1:
                h = self.dropouts[i](h)

        mean = self.last_fc(h)
        #mean = self.output_activation(self.last_fc(h))
        
        if self.std is None:
            log_std = self.last_fc_log_std(h)
            log_std = torch.clamp(log_std, LOG_SIG_MIN, LOG_SIG_MAX)
            std = torch.exp(log_std)
        else:
            std = self.std
            log_std = self.log_std

        log_prob = None
        expected_log_prob = None
        mean_action_log_prob = None
        pre_tanh_value = None
        if deterministic:
            action = torch.tanh(mean)
        else:
            tanh_normal = TanhNormal(mean, std)
            if return_log_prob:
                if reparameterize:
                    action, pre_tanh_value = tanh_normal.rsample(
                        return_pretanh_value=True
                    )
                else:
                    action, pre_tanh_value = tanh_normal.sample(
                        return_pretanh_value=True
                    )
                log_prob = tanh_normal.log_prob(
                    action,
                    pre_tanh_value=pre_tanh_value
                )
                log_prob = log_prob.sum(dim=1, keepdim=True)
            else:
                if reparameterize:
                    action = tanh_normal.rsample()
                else:
                    action = tanh_normal.sample()

        return (
            action, mean, log_std, log_prob, expected_log_prob, std,
            mean_action_log_prob, pre_tanh_value,
        )

class TransformerPolicy(BERT, ExplorationPolicy):

    def __init__(
        self,
        hidden,
        input_size,
        output_size,
        latent_dim,
        n_layers,
        attn_heads,
        use_sequence_attention=True,
        use_multihead_attention=True,
        use_channel_attention=True,
        mode='parallel',
        std=None,
        init_w=1e-3,
        **kwargs
    ):
        self.save_init_params(locals())
        super().__init__(
            hidden=hidden,
            input_size=input_size,
            output_size=output_size,
            n_layers=n_layers,
            attn_heads=attn_heads,
            use_sequence_attention=use_sequence_attention,
            use_multihead_attention=use_multihead_attention,
            use_channel_attention=use_channel_attention,
            mode='parallel',
            dropout=0.1
        )
        self.latent_dim = latent_dim
        self.log_std = None
        self.std = std
        self.mean = None
        self.output_size = output_size
        self.layer_log_std = nn.Linear(hidden, output_size)
        self.layer_log_std.weight.data.uniform_(-init_w, init_w)
        self.layer_log_std.bias.data.uniform_(-init_w, init_w)

    def get_action(self, obs, deterministic=False):
        #print("here 1", obs.shape)
        actions = self.get_actions(obs, deterministic=deterministic)
        return actions[0, :], {}

    @torch.no_grad()
    def get_actions(self, obs, deterministic=False):
        #print("here 2", obs.shape)
        #obs = obs.unsqueeze(0)
        outputs = self.forward(1, obs.size()[0], obs, deterministic=deterministic)[0]
        return np_ify(outputs)

    def forward(self, meta_size, batch_size, obs, reparameterize=False, deterministic=False, return_log_prob=False):
        x = obs
        #t, b, dim = x.size()
        #x = x.view(t*b, 1, dim)
        #x = x.unsqueeze(1)
        x = x.view(meta_size, batch_size, -1)
        x = self.front(x)
        for transformer in self.transformer_blocks:
            x = transformer.forward(x)
        #self.mean = self.tail(x).view(t, b, -1)
        #self.mean = self.tail(x).squeeze(1)
        self.mean = self.tail(x).view(meta_size*batch_size, -1)
        #self.log_std = self.layer_log_std(x).view(t, b, -1)
        #self.log_std = self.layer_log_std(x).squeeze(1)
        self.log_std = self.layer_log_std(x).view(meta_size*batch_size, -1)
        self.log_std = torch.clamp(self.log_std, LOG_SIG_MIN, LOG_SIG_MAX)
        self.std = torch.exp(self.log_std)
        
        log_prob = None
        pre_tanh_value = None
        if deterministic:
            action = torch.tanh(self.mean)
        else:
            tanh_normal = TanhNormal(self.mean, self.std)
            if return_log_prob:
                if reparameterize:
                    action, pre_tanh_value = tanh_normal.rsample(return_pretanh_value=True)
                else:
                    action, pre_tanh_value = tanh_normal.sample( return_pretanh_value=True)
                log_prob = tanh_normal.log_prob(action, pre_tanh_value=pre_tanh_value)
                log_prob = log_prob.sum(dim=-1, keepdim=True)
            else:
                if reparameterize:
                    action = tanh_normal.rsample()
                else:
                    action = tanh_normal.sample()

        return (action, self.mean, self.log_std, log_prob, self.std, pre_tanh_value,)

class MakeDeterministic(Wrapper, Policy):
    def __init__(self, stochastic_policy):
        super().__init__(stochastic_policy)
        self.stochastic_policy = stochastic_policy

    def get_action(self, observation):
        return self.stochastic_policy.get_action(observation, deterministic=True)
    def get_actions(self, observations):
        return self.stochastic_policy.get_actions(observations, deterministic=True)
